﻿TotalRP3_IntendedUI = {}
local ui = TotalRP3_IntendedUI
local IsControlKeyDown, IsShiftKeyDown, IsModifierKeyDown = IsControlKeyDown, IsShiftKeyDown, IsModifierKeyDown

--[[
	properties
		.parent      parent frame
		.top         top coordinate of TOPLEFT anchor
		.left        left coordinate of TOPLEFT anchor
		.height      table height
		.width       table width
		.selectMode  row select mode: none, single, multiple; default none
		.noHeader    whether or not the table should have a header; default false
		.onUpdate    on update handler for table
]]--
ui.CreateTable = function(self, properties, dataSource, columns)
	
	local tableObj = CreateFrame("ScrollFrame", nil, properties.parent)
	tableObj:SetPoint("TOPLEFT", properties.left, -properties.top)
	
	tableObj.rowHeight = 16
	if properties.noHeader then
		tableObj.headerHeight = 0
	else
		tableObj.headerHeight = 16
	end
	tableObj.scrollBarWidth = 16
	tableObj.onUpdate = properties.onUpdate
	
	tableObj.columns = columns
	tableObj.selectMode = properties.selectMode or "none" -- none, single, multiple
	
	local scrollbar = CreateFrame("Slider", nil, tableObj, "UIPanelScrollBarTemplate")
	scrollbar:SetPoint("TOPLEFT", tableObj, "TOPRIGHT", -tableObj.scrollBarWidth, -tableObj.headerHeight)
	scrollbar:SetPoint("BOTTOMRIGHT", 0, 16)
	scrollbar:SetMinMaxValues(0, 1000)
	scrollbar:SetValue(0)
	
	tableObj.Sort = function(self, colIndex)
		if colIndex then
			local p = {}
			local colToSort = self.columns[math.abs(colIndex)].sortBy or self.columns[math.abs(colIndex)].name
			
			local temp = {}
			
			for indexx, item in pairs(self.dataSource) do
				temp[indexx] = {["idx"] = indexx, val = item[colToSort]}
			end
			if colIndex > 0 then
				table.sort(temp, function(v1, v2) 
					if not v1 and not v2 then return false end
					if not v1 then return false end
					if not v2 then return true end
					if not v1.val and not v2.val then return false end
					if not v1.val then return false end
					if not v2.val then return true end
					return v1.val < v2.val
				end)
			else
				table.sort(temp, function(v1, v2) 
					if not v1 and not v2 then return false end
					if not v1 then return false end
					if not v2 then return true end
					if not v1.val and not v2.val then return false end
					if not v1.val then return false end
					if not v2.val then return true end
					return v1.val > v2.val
				end)
			end
			
			for index = 1,#temp do
				p[index] = temp[index].idx
			end
			
			self.permutation = p
			self.colIndex = colIndex
		else
			local p = {}
			for i = 1,#self.dataSource do
				p[i] = i
			end
			self.permutation = p
			self.colIndex = nil
		end
	end
	
	tableObj.GetDataSource = function(self)
		return self.dataSource
	end
	
	tableObj.Update = function(self, dataSource)
		if dataSource then
			self.dataSource = dataSource
			self.lastPickedData = nil
			self:Sort(self.colIndex)
		end
		
		self.count = #self.dataSource
		
		local i = 0
		
		local scrollHeight = self.count * self.rowHeight - self.contentHeight
		if scrollHeight <= 0 then
			self.scrollbar:SetValue(0)
			self.scrollbar:Disable()
		else
			local offset = self.scrollbar:GetValue() * scrollHeight / 1000 / self.rowHeight
			i = math.floor(offset)
			self.content.rows[1]:SetPoint("TOPLEFT", 0, (offset - i) * self.rowHeight)
			self.scrollbar:Enable()
		end
		
		local rowNumber = 1
		while i < self.count and rowNumber <= self.numVisibleRows do
			i = i + 1
			local dataIndex = self.permutation[i]
			self.content.rows[rowNumber]:Show()
			if self.dataSource[dataIndex].selected then
				self.content.rows[rowNumber].bg:SetColorTexture(0.5, 0.5, 0.5, 1)
			else
				self.content.rows[rowNumber].bg:SetTexture(nil)
			end
			
			for colNumber, column in pairs(self.columns) do
				self.content.rows[rowNumber].dataIndex = dataIndex
				local data = self.dataSource[dataIndex][column.name]
				local cell = self.content.rows[rowNumber].col[colNumber]
				if column.type then
					if column.type == "icon" then
						cell.bg:SetTexture(data)
					elseif column.type == "bool" then
						if data then
							cell.checkMark:SetTexture("INTERFACE\\BUTTONS\\UI-CheckBox-Check")
						else
							cell.checkMark:SetTexture()
						end
					elseif column.type == "action" then
						cell.text:SetText(data or column.caption)
					elseif column.type == "button" then
						-- NOP
					else
						cell.text:SetText(data)
					end
				else
					cell.text:SetText(data)
				end
			end
			rowNumber = rowNumber + 1
		end
		while rowNumber <= #self.content.rows do
			self.content.rows[rowNumber].dataIndex = nil
			self.content.rows[rowNumber]:Hide()
			rowNumber = rowNumber + 1
		end
		
		if self.onUpdate then
			self.onUpdate(self)
		end
	end
	
	if tableObj.selectMode == "multiple" then
		tableObj.SetAllSelected = function(self, selected)
			for i = 1, #self.dataSource do
				self.dataSource[i].selected = selected
			end
			self.lastPickedData = #self.dataSource
			self:Update()
		end
	elseif tableObj.selectMode == "single" then
		tableObj.SetSelectedIndex = function(self, index)
			for i = 1, #self.dataSource do
				self.dataSource[i].selected = nil
			end
			if index and self.dataSource[index] then
				self.dataSource[index].selected = true
			end
			self:Update()
		end
		tableObj.GetSelectedIndex = function(self)
			for i = 1, #self.dataSource do
				if self.dataSource[i].selected then
					return i
				end
			end
			return nil
		end
	end
	
	local scrollbg = scrollbar:CreateTexture(nil, "BACKGROUND")
	scrollbg:SetAllPoints(scrollbar)
	scrollbg:SetColorTexture(0, 0, 0, 0.6)
	tableObj.scrollbar = scrollbar

	local content = CreateFrame("Frame", nil, tableObj)
	content:SetPoint("TOPLEFT", 0, -tableObj.headerHeight)
	content:SetPoint("BOTTOMRIGHT", -tableObj.scrollBarWidth, 0)
	content:SetClipsChildren(true)
	local contentbg = content:CreateTexture(nil, "BACKGROUND")
	contentbg:SetAllPoints(content)
	contentbg:SetColorTexture(0, 0, 0, 0.4)
	tableObj.content = content

	tableObj.header = {}
	
	if not properties.noHeader then
		
		local prevCol = nil
		
		for index, column in pairs(columns) do
			local h = CreateFrame("Frame", nil, tableObj)
			h:SetSize(column.width, tableObj.headerHeight)
			if prevCol then
				h:SetPoint("LEFT", prevCol, "RIGHT")
			else
				h:SetPoint("TOPLEFT")
			end
			prevCol = h
			h.text = h:CreateFontString(nil, "ARTWORK", "GameFontWhite")
			h.text:SetJustifyH("LEFT")
			h.text:SetPoint("TOPLEFT", 2, 0)
			h.text:SetPoint("BOTTOMRIGHT", -2, 0)
			h.text:SetText(column.caption)
			h.index = index
			
			h.bg = h:CreateTexture(nil, "BACKGROUND")
			h.bg:SetAllPoints(h)
			h.bg:SetColorTexture(0, 0, 0, 0.8)
			
			if column.sortable then		
				h:SetScript("OnMouseUp", function(self, button)
					if h:GetParent().colIndex == self.index then
						h:GetParent():Sort(-self.index)
					else
						h:GetParent():Sort(self.index)
					end
					h:GetParent():Update()
				end)
			end
			
			tableObj.header[index] = h
		end
	end
	
	content.rows = {}
	
	tableObj.Resize = function(self, width, height)
		self.contentHeight = height - self.headerHeight
		self:EnsureCapacity(height)
		self:SetSize(width, height)
		
		local totalWidth = 0
		local totalGrow = 0
		local totalShrink = 0
		for i = 1, #self.columns do
			totalWidth = totalWidth + self.columns[i].width
			totalGrow = totalGrow + (self.columns[i].grow or 0)
			totalShrink = totalShrink + (self.columns[i].shrink or 0)
		end
		
		local widthToDistribute = width - self.scrollBarWidth - totalWidth
		
		local widths = {}
		local field
		local factor = 0
		if widthToDistribute < 0 then
			field = "shrink"
			if totalShrink > 0 then
				factor = 1/totalShrink
			end
		else
			field = "grow"
			if totalGrow > 0 then
				factor = 1/totalGrow
			end
		end
				
		for i = 1, #self.columns do
			widths[i] = self.columns[i].width + (self.columns[i][field] or 0) * factor * widthToDistribute
		end
		
		for h = 1, #self.header do
			self.header[h]:SetWidth(widths[h])
		end
		
		for r = 1, #self.content.rows do
			for c = 1, #self.columns do
				self.content.rows[r].col[c]:SetWidth(widths[c])
			end
		end
		
		self:Update(self.dataSource)
	end
	
	tableObj.EnsureCapacity = function(self, height)
		self.numVisibleRows = math.ceil(self.contentHeight / self.rowHeight) + 1
		local prevRow = self.content.rows[#self.content.rows]
		for i = #self.content.rows + 1, self.numVisibleRows do
			local row = CreateFrame("Frame", nil, self.content)
			if prevRow then
				row:SetPoint("TOPLEFT", prevRow, "BOTTOMLEFT")
				row:SetPoint("TOPRIGHT", prevRow, "BOTTOMRIGHT")
			else
				row:SetPoint("TOPLEFT")
				row:SetPoint("RIGHT")
			end
			prevRow = row
			row:SetHeight(self.rowHeight)
			row.bg = row:CreateTexture(nil, "BACKGROUND")
			row.bg:SetAllPoints(row)
			local offset = 0
			row.col = {}
			local prevCol = nil
			for colIndex, col in pairs(columns) do
				local c = CreateFrame("Frame", nil, row)
				if prevCol then
					c:SetPoint("LEFT", prevCol, "RIGHT")
				else
					c:SetPoint("LEFT")
				end
				prevCol = c
				c:SetSize(col.width, self.rowHeight)
				if col.type and col.type == "icon" then
					c.bg = c:CreateTexture(nil, "BACKGROUND")
					c.bg:SetAllPoints(c)
				elseif col.type and col.type == "bool" then
					c.bg = c:CreateTexture(nil, "ARTWORK")
					c.bg:SetPoint("CENTER")
					c.bg:SetSize(16, 16)
					c.bg:SetTexture("INTERFACE\\BUTTONS\\UI-CheckBox-Up")
					c.bg:SetTexCoord(0.125, 0.875, 0.125, 0.875) -- l r t b
					c.bg:SetAlpha(1)
					c.checkMark = c:CreateTexture(nil, "OVERLAY")
					c.checkMark:SetPoint("CENTER")
					c.checkMark:SetSize(16, 16)
					c.checkMark:SetAlpha(1)
					c.checkMark:SetTexCoord(0.125, 0.875, 0.125, 0.875) -- l r t b
				elseif col.type and col.type == "button" then
					c.button = CreateFrame("Button", nil, c, col.template)
					c.button:SetScript("OnClick", function(self, button) 
						local dataIndex = self:GetParent():GetParent().dataIndex
						if dataIndex then
							col.onClick(self:GetParent():GetParent():GetParent():GetParent(), dataIndex)
						end
					end)
				else
					--c.text = c:CreateFontString(nil, "ARTWORK", "GameFontWhite")
					c.text = c:CreateFontString(nil, "ARTWORK", "NumberFontNormal")
					c.text:SetJustifyH(col.align or "LEFT")
					c.text:SetPoint("TOPLEFT", 2, 0)
					c.text:SetPoint("BOTTOMRIGHT", -2, 0)
					if col.type and col.type == "link" then
						c.text:SetTextColor(0.5,0.5,1)
					elseif col.type and col.type == "action" then
						c.text:SetTextColor(1.0,0.75,0.25)
					end			
				end
				
				if tableObj.selectMode == "multiple" then
					c:SetScript("OnMouseUp", function(self, button) 
						local dataIndex = self:GetParent().dataIndex
						if dataIndex then
							local tableObj = self:GetParent():GetParent():GetParent()
							if IsShiftKeyDown() and tableObj.lastPickedData then
								local thisIndex, lastIndex, minIndex, maxIndex
								for displayRow = 1, #tableObj.permutation do
									local dataRow = tableObj.permutation[displayRow]
									if tableObj.dataSource[dataRow].selected then
										minIndex = minIndex or displayRow
										maxIndex = displayRow
									end
									if dataRow == dataIndex then
										thisIndex = displayRow
									elseif dataRow == tableObj.lastPickedData then
										lastIndex = displayRow
									end	
								end
								local startIndex, endIndex
								if lastIndex == minIndex then
									lastIndex = maxIndex
								else
									lastIndex = minIndex
								end
								for i = 1, math.min(thisIndex, lastIndex)-1 do
									tableObj.dataSource[tableObj.permutation[i]].selected = nil
								end
								for i = math.min(thisIndex, lastIndex), math.max(thisIndex, lastIndex) do
									tableObj.dataSource[tableObj.permutation[i]].selected = true
								end
								for i = math.max(thisIndex, lastIndex)+1, #tableObj.permutation do
									tableObj.dataSource[tableObj.permutation[i]].selected = nil
								end
								tableObj.lastPickedData = dataIndex
								tableObj:Update()
							elseif IsControlKeyDown() then
								tableObj.dataSource[dataIndex].selected = not tableObj.dataSource[dataIndex].selected
								tableObj.lastPickedData = nil
								tableObj:Update()
							elseif not col.onMouseUp or IsShiftKeyDown() then
								for i = 1, #tableObj.dataSource do
									tableObj.dataSource[i].selected = nil
								end
								tableObj.dataSource[dataIndex].selected = true
								tableObj.lastPickedData = dataIndex
								tableObj:Update()
							else
								col.onMouseUp(tableObj, dataIndex, button)
							end
						end
					end)
				elseif tableObj.selectMode == "single" then
					c:SetScript("OnMouseUp", function(self, button) 
						local dataIndex = self:GetParent().dataIndex
						if dataIndex then
							if col.onMouseUp and not IsModifierKeyDown() then
								col.onMouseUp(self:GetParent():GetParent():GetParent(), dataIndex, button)
							else
								self:GetParent():GetParent():GetParent():SetSelectedIndex(dataIndex)
							end
						end
					end)
				elseif col.onMouseUp then
					c:SetScript("OnMouseUp", function(self, button) 
						local dataIndex = self:GetParent().dataIndex
						if dataIndex then
							col.onMouseUp(self:GetParent():GetParent():GetParent(), dataIndex, button)
						end
					end)
				end
				
				if col.onEnter then
					c:SetScript("OnEnter", function(self) 
						if self:GetParent().dataIndex then
							col.onEnter(self:GetParent():GetParent():GetParent(), self:GetParent().dataIndex)
						end
					end)
				end
				if col.onLeave then
					c:SetScript("OnLeave", function(self) 
						if self:GetParent().dataIndex then
							col.onLeave(self:GetParent():GetParent():GetParent(), self:GetParent().dataIndex)
						end
					end)
				end
				row.col[colIndex] = c
				offset = offset + col.width
			end
			row:SetScript("OnMouseWheel", function(self, delta)
				local s = self:GetParent():GetParent()
				local scrollHeight = s.count * s.rowHeight - s.contentHeight
				if scrollHeight > 0 then
					local currPosition = s.scrollbar:GetValue()
					local newPosition = math.max(0, math.min(currPosition - delta * s.rowHeight * 1000 / scrollHeight, 1000))
					if currPosition ~= newPosition then
						s.scrollbar:SetValue(newPosition)
						s:Update()
					end
				end
			end)
			table.insert(self.content.rows, row)
		end
	end
	
	scrollbar:SetScript("OnValueChanged", function (self, value)
		self:GetParent():Update()
	end)
	
	tableObj.dataSource = dataSource
	tableObj:Resize(properties.width, properties.height)
	
	return tableObj
end
